#!/usr/bin/env python
from py_hierarchy_2_2d import *
from py_hierarchy_2_2d import __version__ as v
from py_hierarchy_2_2d.tests.utils import my_tabulate, get_project_dir, slurp
import xmltodict
import json
from os import path

# dev dependency only
import toml

# save some keystrokes
mt = my_tabulate


def test_version():
    "Tests the verion in __init__ against the version in pyproject.toml."
    pyproject = toml.load(path.join(get_project_dir(), "pyproject.toml"))
    pyproject_version = pyproject["tool"]["poetry"]["version"]
    assert v == pyproject_version, "version doesn't match pyproject verison?"


def test_airplanes():
    xml = slurp(path.join(get_project_dir(), "sample-data/airplanes.xml"))

    actual = mt(parse(xml), no_print=True)
    expected = """
parent_path    path              tag     attributes    value
-------------  ----------------  ------  ------------  -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
/              /planes/          planes  {}            {}
/planes/       /planes/plane/0/  plane   {}            {'make': 'Cessna', 'year': '1977', 'model': 'MoonDart', 'at-color': 'blue', 'owner': ['mitch', 'glen', 'paulie'], 'action': ['sample text - landed the plane and it was windy.', 'sample text - landed the plane and it was sunny.'], '@action.type': ['landing', 'landing'], '@action.weather': ['windy', 'sunny'], '@action.whitespace_attribute': 'whitespace     '}
/planes/       /planes/plane/1/  plane   {}            {'make': 'Cessna', 'year': '1933', 'model': 'Skyhawk', '@model.color': 'blue', 'owner': 'jim', 'action': 'text!', '@action.type': 'takeoff', '@action.weather': 'windy'}
/planes/       /planes/plane/2/  plane   {}            {'make': 'whitespace->>', 'year': '1833', 'model': 'puppy', '@model.color': 'blue', 'owner': 'hal', 'action': 'more with whitespace', '@action.type': 'takeoff', '@action.weather': 'blowy'}
""".strip()

    assert actual == expected, "Small xml didn't work?"


def test_airplanes_nested_maps():
    xml = slurp(path.join(get_project_dir(), "sample-data/airplanes-nested-maps.xml"))
    actual = mt(parse(xml), no_print=True)
    expected = """
parent_path                   path                               tag          attributes    value
----------------------------  ---------------------------------  -----------  ------------  -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
/                             /planes/                           planes       {}            {}
/planes/                      /planes/plane/0/                   plane        {}            {'make': 'Cessna', 'year': '1977', 'model': 'MoonDart', 'at-color': 'blue', 'owner': ['mitch', 'glen', 'paulie'], 'action': ['sample text - landed the plane and it was windy.', 'sample text - landed the plane and it was sunny.'], '@action.type': ['landing', 'landing'], '@action.weather': ['windy', 'sunny'], '@action.whitespace_attribute': 'whitespace     '}
/planes/plane/0/              /planes/plane/0/plane_trips/       plane_trips  {}            {'trip': ['Mexico', 'Dallas', 'Far Far Away']}
/planes/plane/0/              /planes/plane/0/maintenance/       maintenance  {}            {}
/planes/plane/0/maintenance/  /planes/plane/0/maintenance/fuel/  fuel         {}            {'when': 'monday', 'type': 'gas', 'station': 'shell'}
/planes/                      /planes/plane/1/                   plane        {}            {'make': 'Cessna', 'year': '1933', 'model': 'Skyhawk', '@model.color': 'blue', 'owner': 'jim', 'action': 'text!', '@action.type': 'takeoff', '@action.weather': 'windy'}
/planes/                      /planes/plane/2/                   plane        {}            {'make': 'whitespace->>', 'year': '1833', 'model': 'puppy', '@model.color': 'blue', 'owner': 'hal', 'action': 'more with whitespace', '@action.type': 'takeoff', '@action.weather': 'blowy'}
""".strip()
    assert actual == expected, "Nested maps issues?"


def test_gross_tag_mid_text():
    "TODO"
    tucker_x = """<a>gross-text
                    <b grossAttrib="ugh">gross tag mid-text</b>
                </a>"""

    actual = mt(parse(tucker_x), no_print=True)
    # TODO - decide how to handle this.
    expected = """
parent_path    path    tag    attributes              value
-------------  ------  -----  ----------------------  -----------------------
/              /a/     a      {}                      {'#text': 'gross-text'}
/a/            /a/b/   b      {'grossAttrib': 'ugh'}  gross tag mid-text
""".strip()

    assert actual == expected, "Problem with tags inside text?"


def test_dispatch():
    xml = "<foo><bar>100</bar></foo>"
    a_dict = xmltodict.parse(xml)
    js = json.dumps(a_dict)

    expected = """
parent_path    path    tag    attributes    value
-------------  ------  -----  ------------  --------------
/              /foo/   foo    {}            {'bar': '100'}
""".strip()

    assert expected == mt(parse(xml), no_print=True), "XML dispatch didn't work?"
    assert expected == mt(parse(a_dict), no_print=True), "Dict dispatch didn't work?"
    assert expected == mt(parse(js), no_print=True), "JSON dispatch didn't work?"


def test_nhl():
    teams = {
        "compile-day": "monday",
        "compile-secret": "SECRET!",
        "nhl": [
            {"team": "stars", "players": 10, "pos": ["l", "r", "c"]},
            {"team": "bruins", "players": 30},
            {"team": "preds", "players": 90},
        ],
        "nba": [
            {
                "team": "mavs",
                "players": -9,
                "details": [
                    {"who": "bill", "pos": ["r", "c", "l"]},
                    {"who": "ted", "pos": ["d"]},
                    {"who": "fred"},
                ],
            },
            {"team": "bucks", "players": 3, "details": [{"who": "ken", "pos": ["c"]}]},
        ],
    }

    expected = """
parent_path    path               tag             attributes    value
-------------  -----------------  --------------  ------------  --------------------------------------------------------
/              /compile-day/      compile-day     {}            {'#text': 'monday'}
/              /compile-secret/   compile-secret  {}            SECRET!
/              /nhl/0/            nhl             {}            {'team': 'stars', 'players': 10, 'pos': ['l', 'r', 'c']}
/              /nhl/1/            nhl             {}            {'team': 'bruins', 'players': 30}
/              /nhl/2/            nhl             {}            {'team': 'preds', 'players': 90}
/              /nba/0/            nba             {}            {'team': 'mavs', 'players': -9}
/nba/0/        /nba/0/details/0/  details         {}            {'who': 'bill', 'pos': ['r', 'c', 'l']}
/nba/0/        /nba/0/details/1/  details         {}            {'who': 'ted', 'pos': 'd'}
/nba/0/        /nba/0/details/2/  details         {}            {'who': 'fred'}
/              /nba/1/            nba             {}            {'team': 'bucks', 'players': 3}
/nba/1/        /nba/1/details/0/  details         {}            {'who': 'ken', 'pos': 'c'}
""".strip()
    assert mt(parse(teams), no_print=True) == expected


def test_more_fake_data():
    more = {
        "Coach": {
            "Name": {"Title": "Mr", "Surname": "one", "GivenName": "person"},
            "DOB": "1984-01-01",
            "Gender": "F",
        },
        "TeamPlayers": [
            {
                "Name": {"Title": "Mr", "Surname": "one", "GivenName": "person"},
                "DOB": "1984-01-01",
                "jerseyDetails": [
                    {"JerseyNumber": "#1", "Size": "Large"},
                    {"JerseyNumber": "1", "Size": "Small"},
                ],
                "Gender": "M",
            },
            {
                "Name": {"Title": "Mr", "Surname": "two", "GivenName": "person"},
                "DOB": "1901-01-01",
                "jerseyDetails": [{"JerseyNumber": "ID2", "Size": "Large"}],
                "Gender": "F",
            },
        ],
        "ExtraData": [
            {"Name": "LikesCake", "Value": "true"},
            {"Name": "HasTwin", "Value": "true"},
            {"Name": "EnjoysTwin", "Value": "false"},
            {"Name": "FavCity", "Value": "Dallas"},
            {"Name": "HasChildren", "Value": "false"},
            {"Name": "MathGrade", "Value": "A"},
            {"Name": "US-State", "Value": "CT"},
            {"Name": "Date", "Value": "2019-01-01"},
        ],
    }
    expected = """
parent_path      path                             tag            attributes    value
---------------  -------------------------------  -------------  ------------  --------------------------------------------------------------------------------------------
/                /Coach/                          Coach          {}            {}
/Coach/          /Coach/Name/                     Name           {}            {'Title': 'Mr', 'Surname': 'one', 'GivenName': 'person', 'DOB': '1984-01-01', 'Gender': 'F'}
/                /TeamPlayers/0/                  TeamPlayers    {}            {}
/TeamPlayers/0/  /TeamPlayers/0/Name/             Name           {}            {'Title': 'Mr', 'Surname': 'one', 'GivenName': 'person', 'DOB': '1984-01-01'}
/TeamPlayers/0/  /TeamPlayers/0/jerseyDetails/0/  jerseyDetails  {}            {'JerseyNumber': '#1', 'Size': 'Large'}
/TeamPlayers/0/  /TeamPlayers/0/jerseyDetails/1/  jerseyDetails  {}            {'JerseyNumber': '1', 'Size': 'Small', 'Gender': 'M'}
/                /TeamPlayers/1/                  TeamPlayers    {}            {}
/TeamPlayers/1/  /TeamPlayers/1/Name/             Name           {}            {'Title': 'Mr', 'Surname': 'two', 'GivenName': 'person', 'DOB': '1901-01-01'}
/TeamPlayers/1/  /TeamPlayers/1/jerseyDetails/0/  jerseyDetails  {}            {'JerseyNumber': 'ID2', 'Size': 'Large', 'Gender': 'F'}
/                /ExtraData/0/                    ExtraData      {}            {'Name': 'LikesCake', 'Value': 'true'}
/                /ExtraData/1/                    ExtraData      {}            {'Name': 'HasTwin', 'Value': 'true'}
/                /ExtraData/2/                    ExtraData      {}            {'Name': 'EnjoysTwin', 'Value': 'false'}
/                /ExtraData/3/                    ExtraData      {}            {'Name': 'FavCity', 'Value': 'Dallas'}
/                /ExtraData/4/                    ExtraData      {}            {'Name': 'HasChildren', 'Value': 'false'}
/                /ExtraData/5/                    ExtraData      {}            {'Name': 'MathGrade', 'Value': 'A'}
/                /ExtraData/6/                    ExtraData      {}            {'Name': 'US-State', 'Value': 'CT'}
/                /ExtraData/7/                    ExtraData      {}            {'Name': 'Date', 'Value': '2019-01-01'}
""".strip()

    assert mt(parse(more), no_print=True) != ""



import hashlib
from os import path

# from os import makedirs
from subprocess import getoutput


def if_string_remove_crnl(cell, delim=r"\x01"):
    if not isinstance(cell, str):
        return cell
    for crnl in [
        r"\x00",  # nulls
        delim,  # remove our delim
        "\n",  # unix new lines
        "\r",  # windows carriage returns
    ]:
        cell = cell.replace(crnl, " ")
    return cell


def md5sum(s):
    """take inbound string 's' and return unicode md5 hash (for data uniqueness)

    >>> md5sum('hello world')
    '5eb63bbbe01eeed093cb22bb8f5acdc3'
    """
    m = hashlib.md5()
    m.update(str(s).encode("utf-8"))
    return m.hexdigest()


def spit(f_path, content, fq=False):
    with open(f_path, "w") as i:
        return i.write(content)


def slurp(f_path, fq=False):
    with open(f_path, "r") as i:
        return i.read()


def get_project_dir():
    """Python makes it difficult to yeld our project root... not sure how this
    will work as a python site package; it won't work in service fabric where
    git might not be on the path. For testing only."""
    try:
        return path.abspath(getoutput("git rev-parse --show-toplevel"))
    except FileNotFoundError:
        raise FileNotFoundError("This is not a git repo; Can't find our project dir?")


def prefix_project_dir(relative_path):
    return path.join(get_project_dir(), relative_path)



#!/usr/bin/env python

from tabulate import tabulate
from subprocess import getoutput
from pprint import pprint as pp
from os import path


def my_tabulate(l, no_print=False):
    "This function isn't really important or on the critical path; I just wanted to print dictionaries nicely."
    if isinstance(l, list):
        r = tabulate(l, headers="keys")
    elif isinstance(l, dict):
        r = tabulate([l], headers="keys")
    elif isinstance(l, type(None)):
        r = ""
    else:
        t = type(l)
        raise ValueError("Need a dict or list; got a %s" % t)
    if no_print:
        return r
    else:
        print("")
        print(r)
        print("")
        return None


def my_print_interesting_things(cur, nex):
    print("\n----\nCur is:")
    mt(cur)
    print("----\nNex is:")
    mt(nex)
    print("")


def my_pprint_one_at_a_time(l, print_fn=pp):
    # nl = f'\n{"*" * 80}\n'
    n = "*" * 80
    nl = "\n%s\n" % n
    for x in l:
        print_fn(x)
        input("%sPress Enter to continue...%s" % (nl, nl))
    return None


def get_project_dir():
    """Python makes it difficult to yeld our project root... not sure how this
    will work as a python site package; it probably won't work in serverless as
    git might not be on the path. For testing only?"""
    return path.abspath(getoutput("git rev-parse --show-toplevel"))


def slurp(f):
    with open(f, mode="r") as fi:
        return fi.read()



#!/usr/bin/env python
from functools import reduce, singledispatch
import collections
import json
import xmltodict

OrderedDict = collections.OrderedDict


def handle_leaf(parent_path: str, path: str, tag: str, attributes: dict, value: str):
    return {
        "parent_path": parent_path,
        "path": path,
        "tag": tag,
        "attributes": attributes,
        "value": value,
    }


def handle_dict(parent_path: str, path: str, tag: str, value: OrderedDict):
    assert isinstance(value, (dict, OrderedDict))

    yield handle_leaf(
        parent_path=parent_path,
        path=path,
        tag=tag,
        attributes={k[1:]: v for k, v in value.items() if k.startswith("@")},
        value=value.get("#text", {}),
    )
    yield from walk_tree(value, path)


def handle_list(parent_path: str, path: str, tag: str, value: list):
    assert isinstance(value, list)
    for e in value:
        assert e is None or isinstance(e, (str, dict, OrderedDict))

    for i, e in enumerate(value):
        if isinstance(e, (OrderedDict, dict)):
            _attributes = {k[1:]: v for k, v in e.items() if k.startswith("@")}
            _value = e.get("#text", {})
        elif isinstance(e, str):
            _attributes = {}
            _value = e
        elif e is None:
            _attributes = {}
            _value = ""
        else:
            raise TypeError("found unexpected type")

        yield handle_leaf(
            parent_path=parent_path,
            path="%s%s/" % (path, i),
            tag=tag,
            attributes=_attributes,
            value=_value,
        )

        if isinstance(e, (OrderedDict, dict)):
            yield from walk_tree(e, "%s%s/" % (path, i))


def walk_tree(t, path="/"):
    assert isinstance(t, (dict, OrderedDict))
    for key, value in t.items():
        if key.startswith("@") or key == "#text":  # skip attributes and leaf contents
            continue
        if value is None:  # contents of empty leaf elements, ie <a></a>
            value = ""

        if isinstance(value, (dict, OrderedDict)):
            yield from handle_dict(
                parent_path=path, path="%s%s/" % (path, key), tag=key, value=value
            )

        elif isinstance(value, list):
            yield from handle_list(
                parent_path=path, path="%s%s/" % (path, key), tag=key, value=value
            )

        elif isinstance(value, (str, int)):
            yield handle_leaf(
                parent_path=path,
                path="%s%s/" % (path, key),
                tag=key,
                attributes={},
                value=value,
            )

        else:
            import pdb

            pdb.set_trace()
            raise TypeError("found unexpected type")


def merge_tagvalue_into_row(row, tag, val):
    "Merges a tag:value into a row."
    assert isinstance(row, dict), "Row is not a dict?"
    if tag not in row["value"]:  # it doesn't have our tag yet.
        row["value"][tag] = val
    else:  # It has our tag...
        if isinstance(row["value"][tag], list):  # its a list
            row["value"][tag] = [*row["value"][tag], val]  # append to current list
        else:  # it is not a list, so make it one.
            row["value"][tag] = [row["value"][tag], val]
    return row


def combine_two_rows(cur_row, next_row):
    "Merges the next row's value & attributes into the current row."
    tag = next_row["tag"]
    val = next_row["value"]
    attrib = next_row["attributes"]
    cur_row = merge_tagvalue_into_row(cur_row, tag, val)
    for k, aval in attrib.items():
        atag = "@%s.%s" % (tag, k)
        cur_row = merge_tagvalue_into_row(cur_row, atag, aval)
    return cur_row


def parse_reducer(rows, next_row):
    assert isinstance(rows, (list, type(None))), "rows must be list or none?"
    if not rows:  # so it begins
        return [next_row]  # put it in a list and start.
    if next_row["value"] == {}:  # start over, we found a new row.
        return [*rows, next_row]
    else:  # This is the actual rollup.
        *head, tail = rows
        if not isinstance(tail["value"], (dict, list)):
            # TODO This is the gross nested tag in text that tucker found.
            tail["value"] = {"#text": tail["value"]}
            return [*head, tail, next_row]
        tail = combine_two_rows(tail, next_row)
        return [*head, tail]


def is_maybe_xml(s: str):
    """ (gently) tests if the file is xml by evaluating the first non-whitespace character.
    This is used for disptach to xmltodict and/or json.loads, which in turn will
    throw errors on data which is not well-formed.

    TODO would be great to do this with http content negotiation?"""
    if not isinstance(s, str):
        return False
    if s.lstrip()[0] == "<":
        return True
    else:
        return False


@singledispatch
def parse(data):
    import pdb

    return data


@parse.register(str)
def _(data: str):
    if is_maybe_xml(data):
        # TODO do i need force_cdata?
        # dict_data = xmltodict.parse(data, force_cdata=True)
        dict_data = xmltodict.parse(data)
    else:
        dict_data = json.loads(data, object_hook=OrderedDict)

    intermediate = walk_tree(dict_data)
    return list(reduce(parse_reducer, intermediate, None))


@parse.register(OrderedDict)
@parse.register(dict)
def _(data: dict):
    # intermediate should already be done; they probably already did xmltodict.parse somehow.
    intermediate = walk_tree(data)
    return list(reduce(parse_reducer, intermediate, None))



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['py_hierarchy_2_2d', 'py_hierarchy_2_2d.tests']

package_data = \
{'': ['*']}

install_requires = \
['xmltodict>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'py-hierarchy-2-2d',
    'version': '0.1',
    'description': '',
    'long_description': None,
    'author': 'joefromct',
    'author_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
